@@ -35,8 +35,6 @@
 #define SEVAL_NOLONGJMP 0x040
 
 /* Flags for describe_command, shared between type.def and command.def */
-#define SEVAL_FUNCDEF	0x080		/* only allow function definitions */
-#define SEVAL_ONECMD	0x100		/* only allow a single command */
 #define CDESC_ALL		0x001	/* type -a */
 #define CDESC_SHORTDESC		0x002	/* command -V */
 #define CDESC_REUSABLE		0x004	/* command -v */
@@ -147,6 +145,8 @@ extern void set_bashopts __P((void));
 extern void parse_bashopts __P((char *));
 extern void initialize_bashopts __P((int));
 
+extern void set_compatibility_opts __P((void));
+
 /* Functions from type.def */
 extern int describe_command __P((char *, int));
 
@@ -155,6 +155,7 @@ extern int set_or_show_attributes __P((WORD_LIST *, int, int));
 extern int show_all_var_attributes __P((int, int));
 extern int show_var_attributes __P((SHELL_VAR *, int, int));
 extern int show_name_attributes __P((char *, int));
+extern int show_func_attributes __P((char *, int));
 extern void set_var_attribute __P((char *, int, int));
 
 /* Functions from pushd.def */
@@ -165,6 +166,7 @@ extern WORD_LIST *get_directory_stack __P((int));
 
 /* Functions from evalstring.c */
 extern int parse_and_execute __P((char *, const char *, int));
+extern int evalstring __P((char *, const char *, int));
 extern void parse_and_execute_cleanup __P((void));
 extern int parse_string __P((char *, const char *, int, char **));
 
