@@ -1,24 +1,23 @@
 /* vi_keymap.c -- the keymap for vi_mode in readline (). */
 
-/* Copyright (C) 1987, 1989, 1992 Free Software Foundation, Inc.
+/* Copyright (C) 1987-2009 Free Software Foundation, Inc.
 
-   This file is part of the GNU Readline Library, a library for
-   reading lines of text with interactive input and history editing.
+   This file is part of the GNU Readline Library (Readline), a library
+   for reading lines of text with interactive input and history editing.      
 
-   The GNU Readline Library is free software; you can redistribute it
-   and/or modify it under the terms of the GNU General Public License
-   as published by the Free Software Foundation; either version 2, or
+   Readline is free software: you can redistribute it and/or modify
+   it under the terms of the GNU General Public License as published by
+   the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
 
-   The GNU Readline Library is distributed in the hope that it will be
-   useful, but WITHOUT ANY WARRANTY; without even the implied warranty
-   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+   Readline is distributed in the hope that it will be useful,
+   but WITHOUT ANY WARRANTY; without even the implied warranty of
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
-   The GNU General Public License is often shipped with GNU software, and
-   is generally kept in a file called COPYING or LICENSE.  If you do not
-   have a copy of the license, write to the Free Software Foundation,
-   59 Temple Place, Suite 330, Boston, MA 02111 USA. */
+   You should have received a copy of the GNU General Public License
+   along with Readline.  If not, see <http://www.gnu.org/licenses/>.
+*/
 
 #if !defined (BUFSIZ)
 #include <stdio.h>
@@ -41,7 +40,7 @@ KEYMAP_ENTRY_ARRAY vi_movement_keymap = {
   { ISFUNC, rl_emacs_editing_mode },		/* Control-e */
   { ISFUNC, (rl_command_func_t *)0x0 },		/* Control-f */
   { ISFUNC, rl_abort },				/* Control-g */
-  { ISFUNC, rl_backward },			/* Control-h */
+  { ISFUNC, rl_backward_char },			/* Control-h */
   { ISFUNC, (rl_command_func_t *)0x0 },		/* Control-i */
   { ISFUNC, rl_newline },			/* Control-j */
   { ISFUNC, rl_kill_line },			/* Control-k */
@@ -68,7 +67,7 @@ KEYMAP_ENTRY_ARRAY vi_movement_keymap = {
   { ISFUNC, rl_vi_undo },			/* Control-_ */
 
   /* The start of printing characters. */
-  { ISFUNC, rl_forward },			/* SPACE */
+  { ISFUNC, rl_forward_char },			/* SPACE */
   { ISFUNC, (rl_command_func_t *)0x0 },		/* ! */
   { ISFUNC, (rl_command_func_t *)0x0 },		/* " */
   { ISFUNC, rl_insert_comment },		/* # */
@@ -130,7 +129,7 @@ KEYMAP_ENTRY_ARRAY vi_movement_keymap = {
   { ISFUNC, rl_revert_line },			/* U */
   { ISFUNC, (rl_command_func_t *)0x0 },		/* V */
   { ISFUNC, rl_vi_next_word },			/* W */
-  { ISFUNC, rl_rubout },			/* X */
+  { ISFUNC, rl_vi_rubout },			/* X */
   { ISFUNC, rl_vi_yank_to },			/* Y */
   { ISFUNC, (rl_command_func_t *)0x0 },		/* Z */
 
@@ -150,11 +149,11 @@ KEYMAP_ENTRY_ARRAY vi_movement_keymap = {
   { ISFUNC, rl_vi_end_word },			/* e */
   { ISFUNC, rl_vi_char_search },		/* f */
   { ISFUNC, (rl_command_func_t *)0x0 },		/* g */
-  { ISFUNC, rl_backward },			/* h */
-  { ISFUNC, rl_vi_insertion_mode },		/* i */
+  { ISFUNC, rl_backward_char },			/* h */
+  { ISFUNC, rl_vi_insert_mode },		/* i */
   { ISFUNC, rl_get_next_history },		/* j */
   { ISFUNC, rl_get_previous_history },		/* k */
-  { ISFUNC, rl_forward },			/* l */
+  { ISFUNC, rl_forward_char },			/* l */
   { ISFUNC, rl_vi_set_mark },			/* m */
   { ISFUNC, rl_vi_search_again },		/* n */
   { ISFUNC, (rl_command_func_t *)0x0 },		/* o */
@@ -310,7 +309,6 @@ KEYMAP_ENTRY_ARRAY vi_movement_keymap = {
 #endif /* KEYMAP_SIZE > 128 */
 };
 
-
 KEYMAP_ENTRY_ARRAY vi_insertion_keymap = {
   /* The regular control keys come first. */
   { ISFUNC, (rl_command_func_t *)0x0 },		/* Control-@ */
@@ -327,9 +325,9 @@ KEYMAP_ENTRY_ARRAY vi_insertion_keymap = {
   { ISFUNC, rl_insert },			/* Control-k */
   { ISFUNC, rl_insert },			/* Control-l */
   { ISFUNC, rl_newline },			/* Control-m */
-  { ISFUNC, rl_insert },			/* Control-n */
+  { ISFUNC, rl_menu_complete},			/* Control-n */
   { ISFUNC, rl_insert },			/* Control-o */
-  { ISFUNC, rl_insert },			/* Control-p */
+  { ISFUNC, rl_backward_menu_complete },	/* Control-p */
   { ISFUNC, rl_insert },			/* Control-q */
   { ISFUNC, rl_reverse_search_history },	/* Control-r */
   { ISFUNC, rl_forward_search_history },	/* Control-s */
