@@ -85,7 +85,7 @@ initialize_signames ()
 
   /* POSIX 1003.1b-1993 real time signals, but take care of incomplete
      implementations. Acoording to the standard, both, SIGRTMIN and
-     SIGRTMAX must be defined, SIGRTMIN must be stricly less than
+     SIGRTMAX must be defined, SIGRTMIN must be strictly less than
      SIGRTMAX, and the difference must be at least 7, that is, there
      must be at least eight distinct real time signals. */
 
@@ -192,26 +192,34 @@ initialize_signames ()
 #endif
 
 /* SunOS5 */
-#if defined (SIGLWP)	/* special signal used by thread library */
+#if defined (SIGLWP)	/* Solaris: special signal used by thread library */
   signal_names[SIGLWP] = "SIGLWP";
 #endif
 
-#if defined (SIGFREEZE)	/* special signal used by CPR */
+#if defined (SIGFREEZE)	/* Solaris: special signal used by CPR */
   signal_names[SIGFREEZE] = "SIGFREEZE";
 #endif
 
-#if defined (SIGTHAW)	/* special signal used by CPR */
+#if defined (SIGTHAW)	/* Solaris: special signal used by CPR */
   signal_names[SIGTHAW] = "SIGTHAW";
 #endif
 
-#if defined (SIGCANCEL)	/* thread cancellation signal used by libthread */
+#if defined (SIGCANCEL)	/* Solaris: thread cancellation signal used by libthread */
   signal_names[SIGCANCEL] = "SIGCANCEL";
 #endif
 
-#if defined (SIGXRES)	/* resource control exceeded */
+#if defined (SIGXRES)	/* Solaris: resource control exceeded */
   signal_names[SIGXRES] = "SIGXRES";
 #endif
 
+#if defined (SIGJVM1)	/* Solaris: Java Virtual Machine 1 */
+  signal_names[SIGJVM1] = "SIGJVM1";
+#endif
+
+#if defined (SIGJVM2)	/* Solaris: Java Virtual Machine 2 */
+  signal_names[SIGJVM2] = "SIGJVM2";
+#endif
+
 /* HP-UX */
 #if defined (SIGDIL)	/* DIL signal (?) */
   signal_names[SIGDIL] = "SIGDIL";
